@@ -48,7 +48,7 @@
 #include <osmocom/msc/mncc_int.h>
 #include <osmocom/abis/e1_input.h>
 #include <osmocom/core/bitvec.h>
-#include <osmocom/msc/vlr.h>
+#include <osmocom/vlr/vlr.h>
 #include <osmocom/msc/msub.h>
 #include <osmocom/msc/msc_a.h>
 #include <osmocom/msc/paging.h>
@@ -654,6 +654,20 @@ static int gsm48_cc_rx_setup(struct gsm_trans *trans, struct msgb *msg)
 		gsm48_decode_called(&setup.called,
 			      TLVP_VAL(&tp, GSM48_IE_CALLED_BCD)-1);
 	}
+	/* low layer compatibility */
+	if (TLVP_PRESENT(&tp, GSM48_IE_LOWL_COMPAT) && TLVP_LEN(&tp, GSM48_IE_LOWL_COMPAT) > 0 &&
+	    TLVP_LEN(&tp, GSM48_IE_LOWL_COMPAT) <= sizeof(setup.llc.compat)) {
+		setup.fields |= MNCC_F_LOWL_COMPAT;
+		setup.llc.len = TLVP_LEN(&tp, GSM48_IE_LOWL_COMPAT);
+		memcpy(setup.llc.compat, TLVP_VAL(&tp, GSM48_IE_LOWL_COMPAT), setup.llc.len);
+	}
+	/* high layer compatibility */
+	if (TLVP_PRESENT(&tp, GSM48_IE_HIGHL_COMPAT) && TLVP_LEN(&tp, GSM48_IE_HIGHL_COMPAT) > 0 &&
+	    TLVP_LEN(&tp, GSM48_IE_HIGHL_COMPAT) <= sizeof(setup.hlc.compat)) {
+		setup.fields |= MNCC_F_HIGHL_COMPAT;
+		setup.hlc.len = TLVP_LEN(&tp, GSM48_IE_HIGHL_COMPAT);
+		memcpy(setup.hlc.compat, TLVP_VAL(&tp, GSM48_IE_HIGHL_COMPAT), setup.hlc.len);
+	}
 	/* user-user */
 	if (TLVP_PRESENT(&tp, GSM48_IE_USER_USER)) {
 		setup.fields |= MNCC_F_USERUSER;
@@ -758,7 +772,8 @@ static void rx_mncc_sdp(struct gsm_trans *trans, uint32_t mncc_msg_type, const c
 			const struct gsm_mncc_bearer_cap *bcap)
 {
 	struct codec_filter *codecs = &trans->cc.codecs;
-	struct call_leg *cl = trans->msc_a ? trans->msc_a->cc.call_leg : NULL;
+	struct msc_a *msc_a = trans->msc_a;
+	struct call_leg *cl = msc_a ? msc_a->cc.call_leg : NULL;
 	struct rtp_stream *rtp_cn = cl ? cl->rtp[RTP_TO_CN] : NULL;
 
 	if (sdp[0]) {
@@ -804,6 +819,16 @@ static void rx_mncc_sdp(struct gsm_trans *trans, uint32_t mncc_msg_type, const c
 		return;
 	}
 
+	if (msc_a && osmo_timer_pending(&msc_a->cc.assignment_request_pending)) {
+		/* Still waiting for an Assignment Response.
+		 * For example, when the remote call leg sends some MNCC with SDP with a mismatching codec,
+		 * we start Re-Assignment to match that codec: we send an Assignment Request and wait for a response.
+		 * When we receive another MNCC with SDP from the remote call leg before this Re-Assignment is
+		 * completed, we must not trigger *another* Assignment Request, but instead wait for the Re-Assignment
+		 * to come back with a response first. */
+		return;
+	}
+
 	/* We've already completed Assignment of a voice channel (some time ago), and now the remote side has changed
 	 * to a mismatching codec (list). Try to re-assign this side to a matching codec. */
 	LOG_TRANS(trans, LOGL_INFO, "Remote call leg mismatches assigned codec: %s\n",
@@ -925,10 +950,24 @@ static int gsm48_cc_tx_setup(struct gsm_trans *trans, void *arg)
 			msgb_free(msg);
 			return rc;
 		}
+		rc = bearer_cap_filter_rev_lev(&bearer_cap, trans->vsub->classmark.classmark1.rev_lev);
+		if (rc) {
+			LOG_TRANS(trans, LOGL_ERROR, "No codec offered is supported by phase 1 mobile.\n");
+			trans_free(trans);
+			msgb_free(msg);
+			return rc;
+		}
 		break;
 	case GSM48_BCAP_ITCAP_3k1_AUDIO:
 	case GSM48_BCAP_ITCAP_FAX_G3:
 	case GSM48_BCAP_ITCAP_UNR_DIG_INF:
+		bearer_cap = (struct gsm_mncc_bearer_cap){
+			.transfer = trans->bearer_cap.transfer,
+			.mode = GSM48_BCAP_TMOD_CIRCUIT,
+			.coding = GSM48_BCAP_CODING_GSM_STD,
+			.radio = GSM48_BCAP_RRQ_FR_ONLY,
+		};
+
 		if (csd_bs_list_to_bearer_cap(&bearer_cap, &trans->cc.local.bearer_services) == 0) {
 			LOG_TRANS(trans, LOGL_ERROR, "Error composing Bearer Capability for CC Setup\n");
 
@@ -961,6 +1000,12 @@ static int gsm48_cc_tx_setup(struct gsm_trans *trans, void *arg)
 	/* called party BCD number */
 	if (setup->fields & MNCC_F_CALLED)
 		gsm48_encode_called(msg, &setup->called);
+	/* low layer compatibility */
+	if (setup->fields & MNCC_F_LOWL_COMPAT && setup->llc.len > 0 && setup->llc.len <= sizeof(setup->llc.compat))
+		msgb_tlv_put(msg, GSM48_IE_LOWL_COMPAT, setup->llc.len, setup->llc.compat);
+	/* high layer compatibility */
+	if (setup->fields & MNCC_F_HIGHL_COMPAT && setup->hlc.len > 0 && setup->hlc.len <= sizeof(setup->hlc.compat))
+		msgb_tlv_put(msg, GSM48_IE_HIGHL_COMPAT, setup->hlc.len, setup->hlc.compat);
 	/* user-user */
 	if (setup->fields & MNCC_F_USERUSER)
 		gsm48_encode_useruser(msg, 0, &setup->useruser);
@@ -1105,6 +1150,7 @@ static int gsm48_cc_tx_call_proc_and_assign(struct gsm_trans *trans, void *arg)
 				  trans->bearer_cap.transfer);
 			return -EINVAL;
 		}
+		bearer_cap_filter_rev_lev(&proceeding->bearer_cap, trans->vsub->classmark.classmark1.rev_lev);
 		gsm48_encode_bearer_cap(msg, 0, &proceeding->bearer_cap);
 		memcpy(&trans->bearer_cap, &proceeding->bearer_cap, sizeof(trans->bearer_cap));
 	}
@@ -1820,6 +1866,7 @@ static int gsm48_cc_tx_modify(struct gsm_trans *trans, void *arg)
 	gsm48_start_cc_timer(trans, 0x323, GSM48_T323);
 
 	/* bearer capability */
+	bearer_cap_filter_rev_lev(&modify->bearer_cap, trans->vsub->classmark.classmark1.rev_lev);
 	gsm48_encode_bearer_cap(msg, 1, &modify->bearer_cap);
 	memcpy(&trans->bearer_cap, &modify->bearer_cap, sizeof(trans->bearer_cap));
 
@@ -1867,6 +1914,7 @@ static int gsm48_cc_tx_modify_complete(struct gsm_trans *trans, void *arg)
 	gh->msg_type = GSM48_MT_CC_MODIFY_COMPL;
 
 	/* bearer capability */
+	bearer_cap_filter_rev_lev(&modify->bearer_cap, trans->vsub->classmark.classmark1.rev_lev);
 	gsm48_encode_bearer_cap(msg, 1, &modify->bearer_cap);
 	memcpy(&trans->bearer_cap, &modify->bearer_cap, sizeof(trans->bearer_cap));
 
@@ -1920,6 +1968,7 @@ static int gsm48_cc_tx_modify_reject(struct gsm_trans *trans, void *arg)
 	gh->msg_type = GSM48_MT_CC_MODIFY_REJECT;
 
 	/* bearer capability */
+	bearer_cap_filter_rev_lev(&modify->bearer_cap, trans->vsub->classmark.classmark1.rev_lev);
 	gsm48_encode_bearer_cap(msg, 1, &modify->bearer_cap);
 	memcpy(&trans->bearer_cap, &modify->bearer_cap, sizeof(trans->bearer_cap));
 	/* cause */
@@ -2600,6 +2649,9 @@ int gsm0408_rcv_cc(struct msc_a *msc_a, struct msgb *msg)
 
 	if (!vsub) {
 		LOG_TRANS(trans, LOGL_ERROR, "Invalid conn: no subscriber\n");
+		/* Decrement use counter that has been incremented by CM Service Request (CC).
+		 * If there is no other service request, the BSS connection will be released. */
+		msc_a_put(msc_a, MSC_A_USE_CM_SERVICE_CC);
 		return -EINVAL;
 	}
 
@@ -2617,6 +2669,9 @@ int gsm0408_rcv_cc(struct msc_a *msc_a, struct msgb *msg)
 			rc = gsm48_tx_simple(msc_a,
 					     GSM48_PDISC_CC | (transaction_id << 4),
 					     GSM48_MT_CC_RELEASE_COMPL);
+			/* Decrement use counter that has been incremented by CM Service Request (CC).
+			 * If there is no other service request, the BSS connection will be released. */
+			msc_a_put(msc_a, MSC_A_USE_CM_SERVICE_CC);
 			return -ENOMEM;
 		}
 		if (osmo_fsm_inst_dispatch(msc_a->c.fi, MSC_A_EV_TRANSACTION_ACCEPTED, trans)) {
