@@ -43,6 +43,7 @@
 #include <osmocom/gsm/gsm0411_utils.h>
 #include <osmocom/gsm/protocol/gsm_04_11.h>
 #include <osmocom/gsm/protocol/gsm_03_40.h>
+#include <osmocom/gsm/protocol/gsm_04_08_gprs.h>
 
 #include <osmocom/msc/debug.h>
 #include <osmocom/msc/gsm_data.h>
@@ -52,7 +53,7 @@
 #include <osmocom/msc/signal.h>
 #include <osmocom/msc/db.h>
 #include <osmocom/msc/transaction.h>
-#include <osmocom/msc/vlr.h>
+#include <osmocom/vlr/vlr.h>
 #include <osmocom/msc/msub.h>
 #include <osmocom/msc/msc_a.h>
 #include <osmocom/msc/paging.h>
@@ -231,6 +232,20 @@ static int gsm411_cp_sendmsg(struct msgb *msg, struct gsm_trans *trans,
 	return gsm411_sendmsg(trans, msg);
 }
 
+int gsm411_send_cp_error(struct msc_a *msc_a, uint8_t pdisc, uint8_t cause)
+{
+	struct msgb *msg = gsm411_msgb_alloc();
+	struct gsm48_hdr *gh;
+
+	msgb_put_u8(msg, cause);
+
+	gh = (struct gsm48_hdr *) msgb_push(msg, sizeof(*gh));
+	gh->proto_discr = pdisc;
+	gh->msg_type = GSM411_MT_CP_ERROR;
+
+	return msc_a_tx_dtap_to_i(msc_a, msg);
+}
+
 /* mm_send: receive MMCCSMS sap message from SMC */
 static int gsm411_mm_send(struct gsm411_smc_inst *inst, int msg_type,
 			struct msgb *msg, int cp_msg_type)
@@ -852,7 +867,19 @@ static int gsm411_rx_rp_ack(struct gsm_trans *trans,
 
 	if (trans->net->sms_over_gsup) {
 		/* Forward towards SMSC via GSUP */
-		return gsm411_gsup_mt_fwd_sm_res(trans, rph->msg_ref);
+		uint8_t ui_len = 0;
+		uint8_t *ui_buf = NULL;
+		/* check for:
+		 * - sufficient size for tag, length and data
+		 * - RP-User-Data tag
+		 * - user data length is valid */
+		if (rph->len > 2 &&
+		    rph->data[0] == GSM411_IE_RP_USER_DATA &&
+		    rph->data[1] <= (rph->len - 2)) {
+			ui_len = rph->data[1];
+			ui_buf = &(rph->data[2]);
+		}
+		return gsm411_gsup_mt_fwd_sm_res(trans, rph->msg_ref, ui_buf, ui_len);
 	}
 
 	if (!sms) {
@@ -891,9 +918,34 @@ static int gsm411_rx_rp_error(struct gsm_trans *trans,
 	LOG_TRANS(trans, LOGL_NOTICE, "RX SMS RP-ERROR, cause %d:%d (%s)\n",
 		      cause_len, cause, get_value_string(gsm411_rp_cause_strs, cause));
 
+	if (cause == GSM411_RP_CAUSE_MT_MEM_EXCEEDED)
+		rate_ctr_inc2(net->msc_ctrs, MSC_CTR_SMS_RP_ERR_MEM);
+	else
+		rate_ctr_inc2(net->msc_ctrs, MSC_CTR_SMS_RP_ERR_OTHER);
+
 	if (trans->net->sms_over_gsup) {
 		/* Forward towards SMSC via GSUP */
-		return gsm411_gsup_mt_fwd_sm_err(trans, rph->msg_ref, cause);
+		uint8_t ui_len = 0;
+		uint8_t *ui_buf = NULL;
+		uint8_t ui_tag_idx = cause_len + 1;
+		uint8_t ui_len_idx = ui_tag_idx + 1;
+		uint8_t ui_buf_idx = ui_len_idx + 1;
+		/* check for:
+		 * - sufficient size for tag, length and data
+		 * - RP-User-Data tag
+		 * - user data length is valid */
+		if (rph->len > ui_buf_idx &&
+		    rph->data[ui_tag_idx] == GSM411_IE_RP_USER_DATA &&
+		    rph->data[ui_len_idx] <= (rph->len - ui_buf_idx)) {
+			ui_len = rph->data[ui_len_idx];
+			ui_buf = &(rph->data[ui_buf_idx]);
+		}
+
+		return gsm411_gsup_mt_fwd_sm_err(trans,
+						 rph->msg_ref,
+						 NULL, /* Cause (not indicated by network) */
+						 &cause, /* SM-RP-Cause (indicated by MS/UE) */
+						 ui_buf, ui_len);
 	}
 
 	if (!sms) {
@@ -910,10 +962,8 @@ static int gsm411_rx_rp_error(struct gsm_trans *trans,
 		 * to store this in our database and wait for a SMMA message */
 		/* FIXME */
 		send_signal(S_SMS_MEM_EXCEEDED, trans, sms, 0);
-		rate_ctr_inc(rate_ctr_group_get_ctr(net->msc_ctrs, MSC_CTR_SMS_RP_ERR_MEM));
 	} else {
 		send_signal(S_SMS_UNKNOWN_ERROR, trans, sms, 0);
-		rate_ctr_inc(rate_ctr_group_get_ctr(net->msc_ctrs, MSC_CTR_SMS_RP_ERR_OTHER));
 	}
 
 	sms_free(sms);
@@ -1242,6 +1292,7 @@ int gsm411_send_rp_data(struct gsm_network *net, struct vlr_subscr *vsub,
 {
 	struct gsm_trans *trans;
 	struct msgb *msg;
+	int rc;
 
 	/* Allocate a new transaction for MT SMS */
 	trans = gsm411_alloc_mt_trans(net, vsub);
@@ -1283,9 +1334,13 @@ int gsm411_send_rp_data(struct gsm_network *net, struct vlr_subscr *vsub,
 	/* FIXME: MT SMS is not guaranteed to be delivered (e.g. the MS may be detached) */
 	rate_ctr_inc(rate_ctr_group_get_ctr(net->msc_ctrs, MSC_CTR_SMS_DELIVERED));
 
-	return gsm411_rp_sendmsg(&trans->sms.smr_inst, msg,
-		GSM411_MT_RP_DATA_MT, trans->sms.sm_rp_mr,
-		GSM411_SM_RL_DATA_REQ);
+	rc = gsm411_rp_sendmsg(&trans->sms.smr_inst, msg,
+			       GSM411_MT_RP_DATA_MT,
+			       trans->sms.sm_rp_mr,
+			       GSM411_SM_RL_DATA_REQ);
+	if (rc == 0)
+		trans->sms.gsup_rsp_pending = true;
+	return rc;
 }
 
 /* Entry point for incoming GSM48_PDISC_SMS from abis_rsl.c */
@@ -1311,6 +1366,12 @@ int gsm0411_rcv_sms(struct msc_a *msc_a, struct msgb *msg)
 			"to us anymore. We are ignoring it, maybe a CP-ERROR "
 			"from a MS?\n",
 			transaction_id);
+		/* Decrement use counter that has been incremented by CM Service Request (SMS).
+		 * If there is no other service request, the BSS connection will be released. */
+		if (!osmo_use_count_by(&msc_a->use_count, MSC_A_USE_CM_SERVICE_SMS))
+			LOG_TRANS(trans, LOGL_ERROR, "MO SMS without prior CM Service Request\n");
+		else
+			msc_a_put(msc_a, MSC_A_USE_CM_SERVICE_SMS);
 		return -EINVAL;
 	}
 
@@ -1318,7 +1379,15 @@ int gsm0411_rcv_sms(struct msc_a *msc_a, struct msgb *msg)
 		new_trans = 1;
 		trans = gsm411_trans_init(net, vsub, msc_a, transaction_id, true);
 		if (!trans) {
-			/* FIXME: send some error message */
+			/* Send error message. */
+			gsm411_send_cp_error(msc_a, GSM48_PDISC_SMS | (transaction_id << 4),
+					     GSM411_CP_CAUSE_NET_FAIL);
+			/* Decrement use counter that has been incremented by CM Service Request (SMS).
+			 * If there is no other service request, the BSS connection will be released. */
+			if (!osmo_use_count_by(&msc_a->use_count, MSC_A_USE_CM_SERVICE_SMS))
+				LOG_TRANS(trans, LOGL_ERROR, "MO SMS without prior CM Service Request\n");
+			else
+				msc_a_put(msc_a, MSC_A_USE_CM_SERVICE_SMS);
 			return -ENOMEM;
 		}
 
@@ -1379,6 +1448,17 @@ void _gsm411_sms_trans_free(struct gsm_trans *trans)
 		sms_free(trans->sms.sms);
 		trans->sms.sms = NULL;
 	}
+
+	if (trans->net->sms_over_gsup) {
+		if (trans->sms.gsup_rsp_pending) {
+			const int net_cause = GMM_CAUSE_NET_FAIL;
+			gsm411_gsup_mt_fwd_sm_err(trans,
+						  trans->sms.sm_rp_mr,
+						  &net_cause, /* Cause IE (indicated by network) */
+						  NULL, /* SM-RP-Cause (not indicated by MS/UE) */
+						  NULL, 0 /* SM-RP-UI */);
+		}
+	}
 }
 
 /* Process incoming SAPI N-REJECT from BSC */
