@@ -23,6 +23,7 @@
 #include <osmocom/core/linuxlist.h>
 #include <osmocom/core/logging.h>
 #include <osmocom/core/fsm.h>
+#include <osmocom/core/stat_item.h>
 #include <osmocom/sigtran/sccp_helpers.h>
 
 #include <osmocom/msc/ran_peer.h>
@@ -30,7 +31,7 @@
 #include <osmocom/msc/msub.h>
 #include <osmocom/msc/msc_i.h>
 #include <osmocom/msc/msc_a.h>
-#include <osmocom/msc/vlr.h>
+#include <osmocom/vlr/vlr.h>
 #include <osmocom/msc/ran_conn.h>
 #include <osmocom/msc/cell_id_list.h>
 #include <osmocom/msc/msc_vgcs.h>
@@ -45,6 +46,7 @@ static __attribute__((constructor)) void ran_peer_init()
 /* Allocate a RAN peer with FSM instance. To deallocate, call osmo_fsm_inst_term(ran_peer->fi). */
 static struct ran_peer *ran_peer_alloc(struct sccp_ran_inst *sri, const struct osmo_sccp_addr *peer_addr)
 {
+	struct gsm_network *net = sri->user_data;
 	struct ran_peer *rp;
 	struct osmo_fsm_inst *fi;
 	char *sccp_addr;
@@ -75,6 +77,7 @@ static struct ran_peer *ran_peer_alloc(struct sccp_ran_inst *sri, const struct o
 	fi->priv = rp;
 
 	llist_add(&rp->entry, &sri->ran_peers);
+	osmo_stat_item_inc(osmo_stat_item_group_get_item(net->statg, MSC_STAT_RAN_PEERS_TOTAL), 1);
 
 	return rp;
 }
@@ -87,7 +90,7 @@ struct ran_peer *ran_peer_find_or_create(struct sccp_ran_inst *sri, const struct
 	return ran_peer_alloc(sri, peer_addr);
 }
 
-struct ran_peer *ran_peer_find_by_addr(struct sccp_ran_inst *sri, const struct osmo_sccp_addr *peer_addr)
+struct ran_peer *ran_peer_find_by_addr(const struct sccp_ran_inst *sri, const struct osmo_sccp_addr *peer_addr)
 {
 	struct ran_peer *rp;
 	llist_for_each_entry(rp, &sri->ran_peers, entry) {
@@ -98,6 +101,21 @@ struct ran_peer *ran_peer_find_by_addr(struct sccp_ran_inst *sri, const struct o
 	return NULL;
 }
 
+/* Find a ran_peer (BSC/RNC) by its remote Point Code */
+struct ran_peer *ran_peer_find_by_pc(const struct sccp_ran_inst *sri, uint32_t pc)
+{
+	struct ran_peer *rp;
+	struct osmo_ss7_instance *cs7 = osmo_sccp_get_ss7(sri->sccp);
+	struct osmo_sccp_addr rem_addr;
+
+	osmo_sccp_make_addr_pc_ssn(&rem_addr, pc, sri->ran->ssn);
+	rp = ran_peer_find_by_addr(sri, &rem_addr);
+	if (rp)
+		return rp;
+	LOGP(DMSC, LOGL_DEBUG, "No ran_peer found under remote address: %s\n", osmo_sccp_addr_name(cs7, &rem_addr));
+	return NULL;
+}
+
 void ran_peer_cells_seen_add(struct ran_peer *ran_peer, const struct gsm0808_cell_id *cid)
 {
 	if (!cell_id_list_add_cell(ran_peer, &ran_peer->cells_seen, cid))
@@ -114,7 +132,7 @@ static const struct osmo_tdef_state_timeout ran_peer_fsm_timeouts[32] = {
 #define ran_peer_state_chg(RAN_PEER, NEXT_STATE) \
 	osmo_tdef_fsm_inst_state_chg((RAN_PEER)->fi, NEXT_STATE, ran_peer_fsm_timeouts, g_sccp_tdefs, 5)
 
-void ran_peer_discard_all_conns(struct ran_peer *rp)
+static void ran_peer_discard_all_conns(struct ran_peer *rp)
 {
 	struct ran_conn *conn, *next;
 
@@ -183,7 +201,7 @@ static void ran_peer_rx_reset_ack(struct ran_peer *rp, struct msgb* msg)
 	ran_peer_state_chg(rp, RAN_PEER_ST_READY);
 }
 
-void ran_peer_reset(struct ran_peer *rp)
+static void ran_peer_reset(struct ran_peer *rp)
 {
 	struct msgb *reset;
 
@@ -205,7 +223,7 @@ void ran_peer_reset(struct ran_peer *rp)
 	}
 }
 
-void ran_peer_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+static void ran_peer_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct ran_peer *rp = fi->priv;
 	struct ran_peer_ev_ctx *ctx = data;
@@ -236,7 +254,7 @@ void ran_peer_allstate_action(struct osmo_fsm_inst *fi, uint32_t event, void *da
 	}
 }
 
-void clear_and_disconnect(struct ran_peer *rp, uint32_t conn_id)
+static void clear_and_disconnect(struct ran_peer *rp, uint32_t conn_id)
 {
 	struct msgb *clear;
 	struct ran_msg ran_enc_msg = {
@@ -254,7 +272,7 @@ void clear_and_disconnect(struct ran_peer *rp, uint32_t conn_id)
 	sccp_ran_disconnect(rp->sri, conn_id, 0);
 }
 
-void ran_peer_st_wait_rx_reset(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+static void ran_peer_st_wait_rx_reset(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct ran_peer *rp = fi->priv;
 	struct ran_peer_ev_ctx *ctx;
@@ -292,13 +310,22 @@ void ran_peer_st_wait_rx_reset(struct osmo_fsm_inst *fi, uint32_t event, void *d
 		ran_peer_rx_reset(rp, msg);
 		return;
 
+	case RAN_PEER_EV_AVAILABLE:
+		/* Send a RESET to the peer. */
+		ran_peer_reset(rp);
+		return;
+
+	case RAN_PEER_EV_UNAVAILABLE:
+		/* Do nothing, wait for peer to come up again. */
+		return;
+
 	default:
 		LOG_RAN_PEER(rp, LOGL_ERROR, "Unhandled event: %s\n", osmo_fsm_event_name(&ran_peer_fsm, event));
 		return;
 	}
 }
 
-void ran_peer_st_wait_rx_reset_ack(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+static void ran_peer_st_wait_rx_reset_ack(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct ran_peer *rp = fi->priv;
 	struct ran_peer_ev_ctx *ctx;
@@ -327,6 +354,15 @@ void ran_peer_st_wait_rx_reset_ack(struct osmo_fsm_inst *fi, uint32_t event, voi
 		ran_peer_rx_reset(rp, msg);
 		return;
 
+	case RAN_PEER_EV_AVAILABLE:
+		/* Send a RESET to the peer. */
+		ran_peer_reset(rp);
+		return;
+
+	case RAN_PEER_EV_UNAVAILABLE:
+		ran_peer_state_chg(rp, RAN_PEER_ST_WAIT_RX_RESET);
+		return;
+
 	default:
 		LOG_RAN_PEER(rp, LOGL_ERROR, "Unhandled event: %s\n", osmo_fsm_event_name(&ran_peer_fsm, event));
 		return;
@@ -359,7 +395,15 @@ static struct ran_conn *new_incoming_conn(struct ran_peer *rp, uint32_t conn_id)
 	return msc_i->ran_conn;
 }
 
-void ran_peer_st_ready(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+static void ran_peer_st_ready_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+{
+	struct ran_peer *rp = fi->priv;
+	struct gsm_network *net = rp->sri->user_data;
+	if (prev_state != RAN_PEER_ST_READY)
+		osmo_stat_item_inc(osmo_stat_item_group_get_item(net->statg, MSC_STAT_RAN_PEERS_ACTIVE), 1);
+}
+
+static void ran_peer_st_ready(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct ran_peer *rp = fi->priv;
 	struct ran_peer_ev_ctx *ctx;
@@ -441,12 +485,29 @@ void ran_peer_st_ready(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 		ran_peer_rx_reset(rp, msg);
 		return;
 
+	case RAN_PEER_EV_AVAILABLE:
+		/* Do nothing, we were already up. */
+		return;
+
+	case RAN_PEER_EV_UNAVAILABLE:
+		ran_peer_discard_all_conns(rp);
+		ran_peer_state_chg(rp, RAN_PEER_ST_WAIT_RX_RESET);
+		return;
+
 	default:
 		LOG_RAN_PEER(rp, LOGL_ERROR, "Unhandled event: %s\n", osmo_fsm_event_name(&ran_peer_fsm, event));
 		return;
 	}
 }
 
+static void ran_peer_st_ready_onleave(struct osmo_fsm_inst *fi, uint32_t next_state)
+{
+	struct ran_peer *rp = fi->priv;
+	struct gsm_network *net = rp->sri->user_data;
+	if (next_state != RAN_PEER_ST_READY)
+		osmo_stat_item_dec(osmo_stat_item_group_get_item(net->statg, MSC_STAT_RAN_PEERS_ACTIVE), 1);
+}
+
 static int ran_peer_fsm_timer_cb(struct osmo_fsm_inst *fi)
 {
 	struct ran_peer *rp = fi->priv;
@@ -454,10 +515,17 @@ static int ran_peer_fsm_timer_cb(struct osmo_fsm_inst *fi)
 	return 0;
 }
 
-void ran_peer_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
+static void ran_peer_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
 {
 	struct ran_peer *rp = fi->priv;
+	struct gsm_network *net = rp->sri->user_data;
+
 	ran_peer_discard_all_conns(rp);
+
+	if (rp->fi->state == RAN_PEER_ST_READY)
+		osmo_stat_item_dec(osmo_stat_item_group_get_item(net->statg, MSC_STAT_RAN_PEERS_ACTIVE), 1);
+	osmo_stat_item_dec(osmo_stat_item_group_get_item(net->statg, MSC_STAT_RAN_PEERS_TOTAL), 1);
+
 	llist_del(&rp->entry);
 }
 
@@ -470,8 +538,8 @@ static const struct value_string ran_peer_fsm_event_names[] = {
 	OSMO_VALUE_STRING(RAN_PEER_EV_MSG_DOWN_CO),
 	OSMO_VALUE_STRING(RAN_PEER_EV_RX_RESET),
 	OSMO_VALUE_STRING(RAN_PEER_EV_RX_RESET_ACK),
-	OSMO_VALUE_STRING(RAN_PEER_EV_CONNECTION_SUCCESS),
-	OSMO_VALUE_STRING(RAN_PEER_EV_CONNECTION_TIMEOUT),
+	OSMO_VALUE_STRING(RAN_PEER_EV_AVAILABLE),
+	OSMO_VALUE_STRING(RAN_PEER_EV_UNAVAILABLE),
 	{}
 };
 
@@ -485,7 +553,8 @@ static const struct osmo_fsm_state ran_peer_fsm_states[] = {
 			| S(RAN_PEER_EV_RX_RESET)
 			| S(RAN_PEER_EV_MSG_UP_CO_INITIAL)
 			| S(RAN_PEER_EV_MSG_UP_CO)
-			| S(RAN_PEER_EV_CONNECTION_TIMEOUT)
+			| S(RAN_PEER_EV_AVAILABLE)
+			| S(RAN_PEER_EV_UNAVAILABLE)
 			,
 		.out_state_mask = 0
 			| S(RAN_PEER_ST_WAIT_RX_RESET)
@@ -502,7 +571,8 @@ static const struct osmo_fsm_state ran_peer_fsm_states[] = {
 			| S(RAN_PEER_EV_RX_RESET_ACK)
 			| S(RAN_PEER_EV_MSG_UP_CO_INITIAL)
 			| S(RAN_PEER_EV_MSG_UP_CO)
-			| S(RAN_PEER_EV_CONNECTION_TIMEOUT)
+			| S(RAN_PEER_EV_AVAILABLE)
+			| S(RAN_PEER_EV_UNAVAILABLE)
 			,
 		.out_state_mask = 0
 			| S(RAN_PEER_ST_WAIT_RX_RESET)
@@ -514,6 +584,8 @@ static const struct osmo_fsm_state ran_peer_fsm_states[] = {
 	[RAN_PEER_ST_READY] = {
 		.name = "READY",
 		.action = ran_peer_st_ready,
+		.onenter = ran_peer_st_ready_onenter,
+		.onleave = ran_peer_st_ready_onleave,
 		.in_event_mask = 0
 			| S(RAN_PEER_EV_RX_RESET)
 			| S(RAN_PEER_EV_MSG_UP_CO_INITIAL)
@@ -521,6 +593,8 @@ static const struct osmo_fsm_state ran_peer_fsm_states[] = {
 			| S(RAN_PEER_EV_MSG_DOWN_CO_INITIAL)
 			| S(RAN_PEER_EV_MSG_DOWN_CO)
 			| S(RAN_PEER_EV_MSG_DOWN_CL)
+			| S(RAN_PEER_EV_AVAILABLE)
+			| S(RAN_PEER_EV_UNAVAILABLE)
 			,
 		.out_state_mask = 0
 			| S(RAN_PEER_ST_WAIT_RX_RESET)
@@ -675,6 +749,10 @@ int ran_peer_down_paging(struct ran_peer *rp, const struct gsm0808_cell_id *page
 {
 	struct msgb *l2;
 
+	/* ran_peer is not ready for paging (link not ready). */
+	if (rp->fi->state != RAN_PEER_ST_READY)
+		return 0;
+
 	/* There are also the RAN peers that are configured in the neighbor ident for Handover, but if those aren't
 	 * connected, then we can't Page there. */
 	if (!cell_id_list_find(&rp->cells_seen, page_id, 0, false))
