@@ -32,10 +32,9 @@
 #include <osmocom/gsm/gsm48.h>
 #include <osmocom/gsm/ipa.h>
 #include <osmocom/gsupclient/gsup_client.h>
-#include <osmocom/msc/vlr_sgs.h>
-#include <osmocom/msc/vlr.h>
-#include <osmocom/msc/debug.h>
-#include <osmocom/msc/gsup_client_mux.h>
+#include <osmocom/vlr/vlr_sgs.h>
+#include <osmocom/vlr/vlr.h>
+#include <osmocom/gsupclient/gsup_client_mux.h>
 #include <osmocom/msc/paging.h>
 
 #include <netinet/in.h>
@@ -201,14 +200,28 @@ struct osmo_tdef msc_tdefs_vlr[] = {
 	{ /* terminator */ }
 };
 
+/* 3GPP TS 24.008, table 11.2 Mobility management timers (network-side) */
+struct osmo_tdef sgsn_tdefs_vlr[] = {
+	{ .T = 3312, .default_val = 60, .unit = OSMO_TDEF_M, .desc = "Subscriber expiration timeout" },
+	{ .T = 3350, .default_val = 6, .desc = "Attach/RAU Complete Reallocation procedure" },
+	{ .T = 3360, .default_val = 6, .desc = "Authentication procedure" },
+	{ .T = 3370, .default_val = 6, .desc = "Identification procedure" },
+	{ /* terminator */ }
+};
+
+struct osmo_tdef *vlr_tdefs;
+
 /* This is just a wrapper around the osmo_tdef API.
  * TODO: we should start using osmo_tdef_fsm_inst_state_chg() */
-uint32_t vlr_timer(struct vlr_instance *vlr, uint32_t timer)
+unsigned long vlr_timer_secs(struct vlr_instance *vlr, int cs_timer, int ps_timer)
 {
 	/* NOTE: since we usually do not need more than one instance of the VLR,
 	 * and since libosmocore's osmo_tdef API does not (yet) support dynamic
 	 * configuration, we always use the global instance of msc_tdefs_vlr. */
-	return osmo_tdef_get(msc_tdefs_vlr, timer, OSMO_TDEF_S, 0);
+	if (vlr_is_cs(vlr))
+		return osmo_tdef_get(vlr_tdefs, cs_timer, OSMO_TDEF_S, 0);
+	else
+		return osmo_tdef_get(vlr_tdefs, ps_timer, OSMO_TDEF_S, 0);
 }
 
 /* return static buffer with printable name of VLR subscriber */
@@ -365,7 +378,7 @@ static int vlr_subscr_use_cb(struct osmo_use_count_entry *e, int32_t old_use_cou
 	else
 		level = LOGL_DEBUG;
 
-	LOGPSRC(DREF, level, file, line, "VLR subscr %s %s %s: now used by %s\n",
+	LOGPSRC(g_vlr_log_cat[OSMO_VLR_LOGC_VLR], level, file, line, "VLR subscr %s %s %s: now used by %s\n",
 		vlr_subscr_name(vsub), (e->count - old_use_count) > 0? "+" : "-", e->use,
 		osmo_use_count_name_buf(buf, sizeof(buf), e->use_count));
 
@@ -426,15 +439,19 @@ int vlr_subscr_purge(struct vlr_subscr *vsub)
 	gsup_msg.message_type = OSMO_GSUP_MSGT_PURGE_MS_REQUEST;
 
 	/* provide HLR number in case we know it */
-	gsup_msg.hlr_enc_len = vsub->hlr.len;
-	gsup_msg.hlr_enc = vsub->hlr.buf;
+	if (vsub->hlr.len) {
+		gsup_msg.hlr_enc_len = vsub->hlr.len;
+		gsup_msg.hlr_enc = vsub->hlr.buf;
+	}
+
+	gsup_msg.cn_domain = vlr_is_cs(vsub->vlr) ? OSMO_GSUP_CN_DOMAIN_CS : OSMO_GSUP_CN_DOMAIN_PS;
 
 	return vlr_subscr_tx_gsup_message(vsub, &gsup_msg);
 }
 
 void vlr_subscr_cancel_attach_fsm(struct vlr_subscr *vsub,
 				  enum osmo_fsm_term_cause fsm_cause,
-				  uint8_t gsm48_cause)
+				  enum gsm48_reject_value gsm48_cause)
 {
 	if (!vsub)
 		return;
@@ -452,7 +469,7 @@ void vlr_subscr_free(struct vlr_subscr *vsub)
 {
 	llist_del(&vsub->list);
 	vlr_stat_item_dec(vsub->vlr, VLR_STAT_SUBSCRIBER_COUNT);
-	DEBUGP(DVLR, "freeing VLR subscr %s (max total use count was %d)\n", vlr_subscr_name(vsub),
+	LOGVSUBP(LOGL_DEBUG, vsub, "freeing VLR subscr (max total use count was %d)\n",
 	       vsub->max_total_use_count);
 
 	/* Make sure SGs timer Ts5 is removed */
@@ -476,7 +493,7 @@ int vlr_subscr_alloc_tmsi(struct vlr_subscr *vsub)
 	for (tried = 0; tried < 100; tried++) {
 		rc = osmo_get_rand_id((uint8_t *) &tmsi, sizeof(tmsi));
 		if (rc < 0) {
-			LOGP(DDB, LOGL_ERROR, "osmo_get_rand_id() failed: %s\n", strerror(-rc));
+			LOGVLR(LOGL_ERROR, "osmo_get_rand_id() failed: %s\n", strerror(-rc));
 			return rc;
 		}
 
@@ -484,7 +501,7 @@ int vlr_subscr_alloc_tmsi(struct vlr_subscr *vsub)
 			int16_t nri_v;
 			osmo_tmsi_nri_v_limit_by_ranges(&tmsi, vlr->cfg.nri_ranges, vlr->cfg.nri_bitlen);
 			osmo_tmsi_nri_v_get(&nri_v, tmsi, vlr->cfg.nri_bitlen);
-			LOGP(DVLR, LOGL_DEBUG, "New NRI from range [%s] = 0x%x --> TMSI 0x%08x\n",
+			LOGVLR(LOGL_DEBUG, "New NRI from range [%s] = 0x%x --> TMSI 0x%08x\n",
 			     osmo_nri_ranges_to_str_c(OTC_SELECT, vlr->cfg.nri_ranges), nri_v, tmsi);
 		}
 
@@ -513,8 +530,8 @@ int vlr_subscr_alloc_tmsi(struct vlr_subscr *vsub)
 		return 0;
 	}
 
-	LOGP(DVLR, LOGL_ERROR, "subscr %s: unable to generate valid TMSI"
-	     " after %d tries\n", vlr_subscr_name(vsub), tried);
+	LOGVSUBP(LOGL_ERROR, vsub, "Unable to generate valid TMSI"
+	     " after %d tries\n", tried);
 	return -1;
 }
 
@@ -542,7 +559,7 @@ struct vlr_subscr *_vlr_subscr_find_or_create_by_imsi(struct vlr_instance *vlr,
 		return NULL;
 	vlr_subscr_get_src(vsub, use, file, line);
 	vlr_subscr_set_imsi(vsub, imsi);
-	LOGP(DVLR, LOGL_INFO, "New subscr, IMSI: %s\n", vsub->imsi);
+	LOGVLR(LOGL_INFO, "New subscr, IMSI: %s\n", vsub->imsi);
 	if (created)
 		*created = true;
 	return vsub;
@@ -572,7 +589,7 @@ struct vlr_subscr *_vlr_subscr_find_or_create_by_tmsi(struct vlr_instance *vlr,
 		return NULL;
 	vlr_subscr_get_src(vsub, use, file, line);
 	vsub->tmsi = tmsi;
-	LOGP(DVLR, LOGL_INFO, "New subscr, TMSI: 0x%08x\n", vsub->tmsi);
+	LOGVLR(LOGL_INFO, "New subscr, TMSI: 0x%08x\n", vsub->tmsi);
 	if (created)
 		*created = true;
 	return vsub;
@@ -583,10 +600,9 @@ static void dedup_vsub(struct vlr_subscr *exists, struct vlr_subscr *vsub)
 	struct vlr_instance *vlr = exists->vlr;
 	int i;
 	int j;
-	LOGP(DVLR, LOGL_NOTICE,
-	     "There is an existing subscriber for IMSI %s used by %s, replacing with new VLR subscr: %s used by %s\n",
+	LOGVSUBP(LOGL_NOTICE, vsub,
+	     "There is an existing subscriber for IMSI %s used by %s, replacing with this VLR subscr, used by %s\n",
 	     exists->imsi, osmo_use_count_to_str_c(OTC_SELECT, &exists->use_count),
-	     vlr_subscr_name(vsub),
 	     osmo_use_count_to_str_c(OTC_SELECT, &vsub->use_count));
 
 	/* Take over some state from the previous vsub */
@@ -613,7 +629,7 @@ static void dedup_vsub(struct vlr_subscr *exists, struct vlr_subscr *vsub)
 			 exists->imsi);
 
 	if (vlr->ops.subscr_inval)
-		vlr->ops.subscr_inval(exists->msc_conn_ref, exists);
+		vlr->ops.subscr_inval(exists->msc_conn_ref, exists, VLR_INVAL_REASON_DUPLICATE_SUBSCR);
 	vlr_subscr_free(exists);
 }
 
@@ -636,13 +652,13 @@ void vlr_subscr_set_imsi(struct vlr_subscr *vsub, const char *imsi)
 
 	/* Set the IMSI on the new subscriber, here. */
 	if (OSMO_STRLCPY_ARRAY(vsub->imsi, imsi) >= sizeof(vsub->imsi)) {
-		LOGP(DVLR, LOGL_NOTICE, "IMSI was truncated: full IMSI=%s, truncated IMSI=%s\n",
+		LOGVLR(LOGL_NOTICE, "IMSI was truncated: full IMSI=%s, truncated IMSI=%s\n",
 		       imsi, vsub->imsi);
 		/* XXX Set truncated IMSI anyway, we currently cannot return an error from here. */
 	}
 
 	vsub->id = atoll(vsub->imsi);
-	DEBUGP(DVLR, "set IMSI on subscriber; IMSI=%s id=%llu\n",
+	LOGVLR(LOGL_DEBUG, "set IMSI on subscriber; IMSI=%s id=%llu\n",
 	       vsub->imsi, vsub->id);
 }
 
@@ -651,7 +667,7 @@ void vlr_subscr_set_imei(struct vlr_subscr *vsub, const char *imei)
 	if (!vsub)
 		return;
 	OSMO_STRLCPY_ARRAY(vsub->imei, imei);
-	DEBUGP(DVLR, "set IMEI on subscriber; IMSI=%s IMEI=%s\n",
+	LOGVLR(LOGL_DEBUG, "set IMEI on subscriber; IMSI=%s IMEI=%s\n",
 	       vsub->imsi, vsub->imei);
 }
 
@@ -660,7 +676,7 @@ void vlr_subscr_set_imeisv(struct vlr_subscr *vsub, const char *imeisv)
 	if (!vsub)
 		return;
 	OSMO_STRLCPY_ARRAY(vsub->imeisv, imeisv);
-	DEBUGP(DVLR, "set IMEISV on subscriber; IMSI=%s IMEISV=%s\n",
+	LOGVLR(LOGL_DEBUG, "set IMEISV on subscriber; IMSI=%s IMEISV=%s\n",
 	       vsub->imsi, vsub->imeisv);
 
 	/* Copy IMEISV to IMEI (additional SV digits get cut off) */
@@ -673,7 +689,7 @@ void vlr_subscr_set_msisdn(struct vlr_subscr *vsub, const char *msisdn)
 	if (!vsub)
 		return;
 	OSMO_STRLCPY_ARRAY(vsub->msisdn, msisdn);
-	DEBUGP(DVLR, "set MSISDN on subscriber; IMSI=%s MSISDN=%s\n",
+	LOGVLR(LOGL_DEBUG, "set MSISDN on subscriber; IMSI=%s MSISDN=%s\n",
 	       vsub->imsi, vsub->msisdn);
 }
 
@@ -688,7 +704,7 @@ void vlr_subscr_set_last_used_eutran_plmn_id(struct vlr_subscr *vsub,
 	} else {
 		vsub->sgs.last_eutran_plmn_present = false;
 	}
-	DEBUGP(DVLR, "set Last E-UTRAN PLMN ID on subscriber: %s\n",
+	LOGVLR(LOGL_DEBUG, "set Last E-UTRAN PLMN ID on subscriber: %s\n",
 	       vsub->sgs.last_eutran_plmn_present ?
 	         osmo_plmn_name(&vsub->sgs.last_eutran_plmn) :
 		 "(none)");
@@ -721,7 +737,7 @@ bool vlr_subscr_matches_imei(struct vlr_subscr *vsub, const char *imei)
 int vlr_subscr_changed(struct vlr_subscr *vsub)
 {
 	/* FIXME */
-	LOGP(DVLR, LOGL_ERROR, "Not implemented: %s\n", __func__);
+	LOGVLR(LOGL_ERROR, "Not implemented: %s\n", __func__);
 	return 0;
 }
 
@@ -731,10 +747,10 @@ void vlr_subscr_enable_expire_lu(struct vlr_subscr *vsub)
 
 	/* Mark the subscriber as inactive if it stopped to do periodical location updates. */
 	if (osmo_clock_gettime(CLOCK_MONOTONIC, &now) == 0) {
-		vsub->expire_lu = now.tv_sec + vlr_timer(vsub->vlr, 3212);
+		vsub->expire_lu = now.tv_sec + vlr_timer_secs(vsub->vlr, 3212, 3312);
 	} else {
-		LOGP(DVLR, LOGL_ERROR,
-		     "%s: Could not enable Location Update expiry: unable to read current time\n", vlr_subscr_name(vsub));
+		LOGVSUBP(LOGL_ERROR, vsub,
+		     "Could not enable Location Update expiry: unable to read current time\n");
 		/* Disable LU expiry for this subscriber. This subscriber will only be freed after an explicit IMSI detach. */
 		vsub->expire_lu = VLR_SUBSCRIBER_NO_EXPIRATION;
 	}
@@ -748,14 +764,14 @@ void vlr_subscr_expire_lu(void *data)
 
 	/* Periodic location update might be disabled from the VTY,
 	 * so we shall not expire subscribers until explicit IMSI Detach. */
-	if (!vlr_timer(vlr, 3212))
+	if (!vlr_timer_secs(vlr, 3212, 3312))
 		goto done;
 
 	if (llist_empty(&vlr->subscribers))
 		goto done;
 
 	if (osmo_clock_gettime(CLOCK_MONOTONIC, &now) != 0) {
-		LOGP(DVLR, LOGL_ERROR, "Skipping Location Update expiry: Could not read current time\n");
+		LOGVLR(LOGL_ERROR, "Skipping Location Update expiry: Could not read current time\n");
 		goto done;
 	}
 
@@ -763,7 +779,7 @@ void vlr_subscr_expire_lu(void *data)
 		if (vsub->expire_lu == VLR_SUBSCRIBER_NO_EXPIRATION || vsub->expire_lu > now.tv_sec)
 			continue;
 
-		LOGP(DVLR, LOGL_DEBUG, "%s: Location Update expired\n", vlr_subscr_name(vsub));
+		LOGVLR(LOGL_DEBUG, "%s: Location Update expired\n", vlr_subscr_name(vsub));
 		vlr_rate_ctr_inc(vlr, VLR_CTR_DETACH_BY_T3212);
 		vlr_subscr_detach(vsub);
 	}
@@ -841,20 +857,20 @@ static int vlr_rx_gsup_unknown_imsi(struct vlr_instance *vlr,
 				    const struct osmo_gsup_message *gsup_msg)
 {
 	if (OSMO_GSUP_IS_MSGT_REQUEST(gsup_msg->message_type)) {
-		LOGP(DVLR, LOGL_NOTICE,
+		LOGVLR(LOGL_NOTICE,
 		     "Unknown IMSI %s, discarding GSUP request "
 		     "of type 0x%02x\n",
 		     gsup_msg->imsi, gsup_msg->message_type);
 		gsup_client_mux_tx_error_reply(vlr->gcm, gsup_msg, GMM_CAUSE_IMSI_UNKNOWN);
 	} else if (OSMO_GSUP_IS_MSGT_ERROR(gsup_msg->message_type)) {
-		LOGP(DVLR, LOGL_NOTICE,
+		LOGVLR(LOGL_NOTICE,
 		     "Unknown IMSI %s, discarding GSUP error "
 		     "of type 0x%02x, cause '%s' (%d)\n",
 		     gsup_msg->imsi, gsup_msg->message_type,
 		     get_value_string(gsm48_gmm_cause_names, gsup_msg->cause),
 		     gsup_msg->cause);
 	} else {
-		LOGP(DVLR, LOGL_NOTICE,
+		LOGVLR(LOGL_NOTICE,
 		     "Unknown IMSI %s, discarding GSUP response "
 		     "of type 0x%02x\n",
 		     gsup_msg->imsi, gsup_msg->message_type);
@@ -886,7 +902,7 @@ int vlr_subscr_req_lu(struct vlr_subscr *vsub)
 	vlr_rate_ctr_inc(vsub->vlr, VLR_CTR_GSUP_TX_UL_REQ);
 
 	gsup_msg.message_type = OSMO_GSUP_MSGT_UPDATE_LOCATION_REQUEST;
-	gsup_msg.cn_domain = vsub->vlr->cfg.is_ps ? OSMO_GSUP_CN_DOMAIN_PS : OSMO_GSUP_CN_DOMAIN_CS;
+	gsup_msg.cn_domain = vlr_is_cs(vsub->vlr) ? OSMO_GSUP_CN_DOMAIN_CS : OSMO_GSUP_CN_DOMAIN_PS;
 	rc = vlr_subscr_tx_gsup_message(vsub, &gsup_msg);
 
 	return rc;
@@ -903,7 +919,7 @@ int vlr_subscr_req_sai(struct vlr_subscr *vsub,
 	gsup_msg.message_type = OSMO_GSUP_MSGT_SEND_AUTH_INFO_REQUEST;
 	gsup_msg.auts = auts;
 	gsup_msg.rand = auts_rand;
-	gsup_msg.cn_domain = OSMO_GSUP_CN_DOMAIN_CS;
+	gsup_msg.cn_domain = vlr_is_cs(vsub->vlr) ? OSMO_GSUP_CN_DOMAIN_CS : OSMO_GSUP_CN_DOMAIN_PS;
 
 	return vlr_subscr_tx_gsup_message(vsub, &gsup_msg);
 }
@@ -940,6 +956,7 @@ int vlr_subscr_tx_auth_fail_rep(const struct vlr_subscr *vsub)
 	struct osmo_gsup_message gsup_msg = {
 		.message_class = OSMO_GSUP_MESSAGE_CLASS_SUBSCRIBER_MANAGEMENT,
 		.message_type = OSMO_GSUP_MSGT_AUTH_FAIL_REPORT,
+		.cn_domain = vlr_is_cs(vsub->vlr) ? OSMO_GSUP_CN_DOMAIN_CS : OSMO_GSUP_CN_DOMAIN_PS,
 	};
 
 	vlr_rate_ctr_inc(vsub->vlr, VLR_CTR_GSUP_TX_AUTH_FAIL_REP);
@@ -1025,13 +1042,13 @@ static void vlr_subscr_gsup_insert_data(struct vlr_subscr *vsub,
 		gsm48_decode_bcd_number2(vsub->msisdn, sizeof(vsub->msisdn),
 					 gsup_msg->msisdn_enc,
 					 gsup_msg->msisdn_enc_len, 0);
-		LOGP(DVLR, LOGL_DEBUG, "IMSI:%s has MSISDN:%s\n",
+		LOGVLR(LOGL_DEBUG, "IMSI:%s has MSISDN:%s\n",
 		     vsub->imsi, vsub->msisdn);
 	}
 
 	if (gsup_msg->hlr_enc) {
 		if (gsup_msg->hlr_enc_len > sizeof(vsub->hlr.buf)) {
-			LOGP(DVLR, LOGL_ERROR, "HLR-Number too long (%zu)\n",
+			LOGVLR(LOGL_ERROR, "HLR-Number too long (%zu)\n",
 				gsup_msg->hlr_enc_len);
 			vsub->hlr.len = 0;
 		} else {
@@ -1044,7 +1061,7 @@ static void vlr_subscr_gsup_insert_data(struct vlr_subscr *vsub,
 	if (gsup_msg->pdp_info_compl) {
 		rc = vlr_subscr_pdp_data_clear(vsub);
 		if (rc > 0)
-			LOGP(DVLR, LOGL_INFO, "Cleared existing PDP info\n");
+			LOGVLR(LOGL_INFO, "Cleared existing PDP info\n");
 	}
 
 	for (idx = 0; idx < gsup_msg->num_pdp_infos; idx++) {
@@ -1166,100 +1183,112 @@ static int vlr_subscr_handle_lu_err(struct vlr_subscr *vsub,
 	return 0;
 }
 
-void vlr_gmm_cause_to_mm_cause(enum gsm48_gmm_cause gmm_cause,
-			       enum gsm48_reject_value *gsm48_rej_p)
+enum gsm48_reject_value vlr_gmm_cause_to_reject_cause_domain(enum gsm48_gmm_cause gmm_cause, bool is_cs)
+{
+	enum gsm48_reject_value reject_cause = vlr_gmm_cause_to_reject_cause(gmm_cause);
+	if (is_cs)
+		return vlr_reject_causes_cs(reject_cause);
+	else
+		return vlr_reject_causes_ps(reject_cause);
+}
+
+enum gsm48_reject_value vlr_gmm_cause_to_reject_cause(enum gsm48_gmm_cause gmm_cause)
 {
-	enum gsm48_reject_value gsm48_rej = GSM48_REJECT_NETWORK_FAILURE;
 	switch (gmm_cause) {
 	case GMM_CAUSE_IMSI_UNKNOWN:
-		gsm48_rej = GSM48_REJECT_IMSI_UNKNOWN_IN_HLR;
-		break;
+		return GSM48_REJECT_IMSI_UNKNOWN_IN_HLR;
 	case GMM_CAUSE_ILLEGAL_MS:
-		gsm48_rej = GSM48_REJECT_ILLEGAL_MS;
-		break;
+		return GSM48_REJECT_ILLEGAL_MS;
 	case GMM_CAUSE_IMEI_NOT_ACCEPTED:
-		gsm48_rej = GSM48_REJECT_IMEI_NOT_ACCEPTED;
-		break;
+		return GSM48_REJECT_IMEI_NOT_ACCEPTED;
 	case GMM_CAUSE_ILLEGAL_ME:
-		gsm48_rej = GSM48_REJECT_ILLEGAL_ME;
-		break;
+		return GSM48_REJECT_ILLEGAL_ME;
 	case GMM_CAUSE_GPRS_NOTALLOWED:
-		gsm48_rej = GSM48_REJECT_GPRS_NOT_ALLOWED;
-		break;
+		return GSM48_REJECT_GPRS_NOT_ALLOWED;
 	case GMM_CAUSE_GPRS_OTHER_NOTALLOWED:
-		gsm48_rej = GSM48_REJECT_SERVICES_NOT_ALLOWED;
-		break;
+		return GSM48_REJECT_SERVICES_NOT_ALLOWED;
 	case GMM_CAUSE_MS_ID_NOT_DERIVED:
-		gsm48_rej = GSM48_REJECT_MS_IDENTITY_NOT_DERVIVABLE;
-		break;
+		return GSM48_REJECT_MS_IDENTITY_NOT_DERVIVABLE;
 	case GMM_CAUSE_IMPL_DETACHED:
-		gsm48_rej = GSM48_REJECT_IMPLICITLY_DETACHED;
-		break;
+		return GSM48_REJECT_IMPLICITLY_DETACHED;
 	case GMM_CAUSE_PLMN_NOTALLOWED:
-		gsm48_rej = GSM48_REJECT_PLMN_NOT_ALLOWED;
-		break;
+		return GSM48_REJECT_PLMN_NOT_ALLOWED;
 	case GMM_CAUSE_LA_NOTALLOWED:
-		gsm48_rej = GSM48_REJECT_LOC_NOT_ALLOWED;
-		break;
+		return GSM48_REJECT_LOC_NOT_ALLOWED;
 	case GMM_CAUSE_ROAMING_NOTALLOWED:
-		gsm48_rej = GSM48_REJECT_ROAMING_NOT_ALLOWED;
-		break;
+		return GSM48_REJECT_ROAMING_NOT_ALLOWED;
 	case GMM_CAUSE_NO_GPRS_PLMN:
-		gsm48_rej = GSM48_REJECT_GPRS_NOT_ALLOWED_IN_PLMN;
-		break;
+		return GSM48_REJECT_GPRS_NOT_ALLOWED_IN_PLMN;
 	case GMM_CAUSE_MSC_TEMP_NOTREACH:
-		gsm48_rej = GSM48_REJECT_MSC_TMP_NOT_REACHABLE;
-		break;
+		return GSM48_REJECT_MSC_TMP_NOT_REACHABLE;
 	case GMM_CAUSE_SYNC_FAIL:
-		gsm48_rej = GSM48_REJECT_SYNCH_FAILURE;
-		break;
+		return GSM48_REJECT_SYNCH_FAILURE;
 	case GMM_CAUSE_CONGESTION:
-		gsm48_rej = GSM48_REJECT_CONGESTION;
-		break;
+		return GSM48_REJECT_CONGESTION;
 	case GMM_CAUSE_SEM_INCORR_MSG:
-		gsm48_rej = GSM48_REJECT_INCORRECT_MESSAGE;
-		break;
+		return GSM48_REJECT_INCORRECT_MESSAGE;
 	case GMM_CAUSE_INV_MAND_INFO:
-		gsm48_rej = GSM48_REJECT_INVALID_MANDANTORY_INF;
-		break;
+		return GSM48_REJECT_INVALID_MANDANTORY_INF;
 	case GMM_CAUSE_MSGT_NOTEXIST_NOTIMPL:
-		gsm48_rej = GSM48_REJECT_MSG_TYPE_NOT_IMPLEMENTED;
-		break;
+		return GSM48_REJECT_MSG_TYPE_NOT_IMPLEMENTED;
 	case GMM_CAUSE_MSGT_INCOMP_P_STATE:
-		gsm48_rej = GSM48_REJECT_MSG_TYPE_NOT_COMPATIBLE;
-		break;
+		return GSM48_REJECT_MSG_TYPE_NOT_COMPATIBLE;
 	case GMM_CAUSE_IE_NOTEXIST_NOTIMPL:
-		gsm48_rej = GSM48_REJECT_INF_ELEME_NOT_IMPLEMENTED;
-		break;
+		return GSM48_REJECT_INF_ELEME_NOT_IMPLEMENTED;
 	case GMM_CAUSE_COND_IE_ERR:
-		gsm48_rej = GSM48_REJECT_CONDTIONAL_IE_ERROR;
-		break;
+		return GSM48_REJECT_CONDTIONAL_IE_ERROR;
 	case GMM_CAUSE_MSG_INCOMP_P_STATE:
-		gsm48_rej = GSM48_REJECT_MSG_NOT_COMPATIBLE;
-		break;
+		return GSM48_REJECT_MSG_NOT_COMPATIBLE;
 	case GMM_CAUSE_PROTO_ERR_UNSPEC:
-		gsm48_rej = GSM48_REJECT_PROTOCOL_ERROR;
-		break;
-
+		return GSM48_REJECT_PROTOCOL_ERROR;
 	case GMM_CAUSE_NO_SUIT_CELL_IN_LA:
+		return GSM48_REJECT_NO_SUIT_CELL_IN_LA;
 	case GMM_CAUSE_MAC_FAIL:
+		return GSM48_REJECT_MAC_FAILURE;
 	case GMM_CAUSE_GSM_AUTH_UNACCEPT:
+		return GSM48_REJECT_GSM_AUTH_UNACCEPTABLE;
 	case GMM_CAUSE_NOT_AUTH_FOR_CSG:
+		return GSM48_REJECT_NOT_AUTH_FOR_CSG;
 	case GMM_CAUSE_SMS_VIA_GPRS_IN_RA:
+		return GSM48_REJECT_SMS_PROV_VIA_GPRS_IN_RA;
 	case GMM_CAUSE_NO_PDP_ACTIVATED:
+		return GSM48_REJECT_NO_PDP_CONTEXT_ACTIVATED;
 	case GMM_CAUSE_NET_FAIL:
-		gsm48_rej = GSM48_REJECT_NETWORK_FAILURE;
-		break;
+		return GSM48_REJECT_NETWORK_FAILURE;
+	default:
+		return GSM48_REJECT_NETWORK_FAILURE;
 	}
+}
 
-	*gsm48_rej_p = gsm48_rej;
+enum gsm48_reject_value vlr_reject_causes_ps(enum gsm48_reject_value reject_cause)
+{
+	switch (reject_cause) {
+	case GSM48_REJECT_CALL_CAN_NOT_BE_IDENTIFIED:
+		return GSM48_REJECT_NETWORK_FAILURE;
+	default:
+		return reject_cause;
+	}
+}
+
+enum gsm48_reject_value vlr_reject_causes_cs(enum gsm48_reject_value reject_cause)
+{
+	switch (reject_cause) {
+	case GSM48_REJECT_NO_SUIT_CELL_IN_LA:
+	case GSM48_REJECT_MAC_FAILURE:
+	case GSM48_REJECT_GSM_AUTH_UNACCEPTABLE:
+	case GSM48_REJECT_NOT_AUTH_FOR_CSG:
+	case GSM48_REJECT_SMS_PROV_VIA_GPRS_IN_RA:
+	case GSM48_REJECT_NO_PDP_CONTEXT_ACTIVATED:
+		return GSM48_REJECT_NETWORK_FAILURE;
+	default:
+		return reject_cause;
+	}
 }
 
 /* Handle LOCATION CANCEL request from HLR */
 static int vlr_subscr_handle_cancel_req(struct vlr_subscr *vsub,
 					const struct osmo_gsup_message *gsup_msg)
 {
-	enum gsm48_reject_value gsm48_rej;
 	enum osmo_fsm_term_cause fsm_cause = OSMO_FSM_TERM_ERROR;
 	struct osmo_gsup_message gsup_reply = {0};
 	int rc, is_update_procedure = !gsup_msg->cancel_type ||
@@ -1272,10 +1301,10 @@ static int vlr_subscr_handle_cancel_req(struct vlr_subscr *vsub,
 		 "update procedure" : "subscription withdraw");
 
 	gsup_reply.message_type = OSMO_GSUP_MSGT_LOCATION_CANCEL_RESULT;
+	gsup_reply.cn_domain = vlr_is_cs(vsub->vlr) ? OSMO_GSUP_CN_DOMAIN_CS : OSMO_GSUP_CN_DOMAIN_PS;
 	rc = vlr_subscr_tx_gsup_message(vsub, &gsup_reply);
 
-	vlr_gmm_cause_to_mm_cause(gsup_msg->cause, &gsm48_rej);
-	vlr_subscr_cancel_attach_fsm(vsub, fsm_cause, gsm48_rej);
+	vlr_subscr_cancel_attach_fsm(vsub, fsm_cause, vlr_gmm_cause_to_reject_cause_domain(gsup_msg->cause, true));
 
 	vlr_rate_ctr_inc(vsub->vlr, VLR_CTR_DETACH_BY_CANCEL);
 	vlr_subscr_detach(vsub);
@@ -1293,7 +1322,7 @@ static int vlr_subscr_handle_check_imei(struct vlr_subscr *vsub, const struct os
 	}
 
 	/* Dispatch result to vsub->lu_fsm, which will either handle the result by itself (Check IMEI early) or dispatch
-         * it further to lu_compl_vlr_fsm (Check IMEI after LU). */
+	 * it further to lu_compl_vlr_fsm (Check IMEI after LU). */
 	if (gsup->message_type == OSMO_GSUP_MSGT_CHECK_IMEI_RESULT) {
 		if (gsup->imei_result == OSMO_GSUP_IMEI_RESULT_ACK)
 			osmo_fsm_inst_dispatch(vsub->lu_fsm, VLR_ULA_E_HLR_IMEI_ACK, NULL);
@@ -1467,9 +1496,14 @@ bool vlr_subscr_expire(struct vlr_subscr *vsub)
 
 static int vlr_subscr_detach(struct vlr_subscr *vsub)
 {
+	int rc = 0;
+
 	/* paranoia: should any LU or PARQ FSMs still be running, stop them. */
 	vlr_subscr_cancel_attach_fsm(vsub, OSMO_FSM_TERM_ERROR, GSM48_REJECT_CONGESTION);
 
+	if (!vsub->imsi_detached_flag)
+		rc = vlr_subscr_purge(vsub);
+
 	vsub->imsi_detached_flag = true;
 	vsub->expire_lu = VLR_SUBSCRIBER_NO_EXPIRATION;
 
@@ -1478,7 +1512,7 @@ static int vlr_subscr_detach(struct vlr_subscr *vsub)
 
 	vlr_subscr_expire(vsub);
 
-	return 0;
+	return rc;
 }
 
 /* See TS 23.012 version 9.10.0 4.3.2.1 "Process Detach_IMSI_VLR" */
@@ -1498,7 +1532,7 @@ void vlr_ran_conn_timeout(struct vlr_subscr *vsub)
 	vlr_subscr_cancel_attach_fsm(vsub, OSMO_FSM_TERM_TIMEOUT, GSM48_REJECT_CONGESTION);
 }
 
-struct vlr_instance *vlr_alloc(void *ctx, const struct vlr_ops *ops)
+struct vlr_instance *vlr_alloc(void *ctx, const struct vlr_ops *ops, bool is_ps)
 {
 	struct vlr_instance *vlr = talloc_zero(ctx, struct vlr_instance);
 	OSMO_ASSERT(vlr);
@@ -1522,6 +1556,7 @@ struct vlr_instance *vlr_alloc(void *ctx, const struct vlr_ops *ops)
 	memcpy(&vlr->ops, ops, sizeof(vlr->ops));
 
 	/* defaults */
+	vlr->cfg.is_ps = is_ps;
 	vlr->cfg.assign_tmsi = true;
 	vlr->cfg.nri_bitlen = OSMO_NRI_BITLEN_DEFAULT;
 	vlr->cfg.nri_ranges = osmo_nri_ranges_alloc(vlr);
@@ -1536,16 +1571,23 @@ struct vlr_instance *vlr_alloc(void *ctx, const struct vlr_ops *ops)
 
 	/* reset shared timer definitions */
 	osmo_tdefs_reset(msc_tdefs_vlr);
+	osmo_tdefs_reset(sgsn_tdefs_vlr);
 
 	/* osmo_auth_fsm.c */
-	OSMO_ASSERT(osmo_fsm_register(&vlr_auth_fsm) == 0);
+	vlr_auth_fsm_init(is_ps);
+
 	/* osmo_lu_fsm.c */
-	vlr_lu_fsm_init();
+	vlr_lu_fsm_init(is_ps);
 	/* vlr_access_request_fsm.c */
-	vlr_parq_fsm_init();
+	vlr_parq_fsm_init(is_ps);
 	/* vlr_sgs_fsm.c */
 	vlr_sgs_fsm_init();
 
+	if (is_ps)
+		vlr_tdefs = sgsn_tdefs_vlr;
+	else
+		vlr_tdefs = msc_tdefs_vlr;
+
 	return vlr;
 
 err_statg:
@@ -1570,13 +1612,15 @@ int vlr_start(struct vlr_instance *vlr, struct gsup_client_mux *gcm)
 	return 0;
 }
 
-/* MSC->VLR: Subscribre has disconnected */
+/* MSC->VLR: Subscriber has disconnected */
 int vlr_subscr_disconnected(struct vlr_subscr *vsub)
 {
 	/* This corresponds to a MAP-ABORT from MSC->VLR on a classic B
 	 * interface */
-	osmo_fsm_inst_term(vsub->lu_fsm, OSMO_FSM_TERM_REQUEST, NULL);
-	osmo_fsm_inst_term(vsub->auth_fsm, OSMO_FSM_TERM_REQUEST, NULL);
+	if (vsub->lu_fsm)
+		osmo_fsm_inst_term(vsub->lu_fsm, OSMO_FSM_TERM_REQUEST, NULL);
+	if (vsub->auth_fsm)
+		osmo_fsm_inst_term(vsub->auth_fsm, OSMO_FSM_TERM_REQUEST, NULL);
 	vsub->msc_conn_ref = NULL;
 
 	return 0;
@@ -1674,3 +1718,26 @@ void log_set_filter_vlr_subscr(struct log_target *target,
 	} else
 		target->filter_map &= ~(1 << LOG_FLT_VLR_SUBSCR);
 }
+
+int g_vlr_log_cat[_OSMO_VLR_LOGC_MAX];
+
+void osmo_vlr_set_log_cat(enum osmo_vlr_cat logc, int logc_num)
+{
+	if (logc < OSMO_VLR_LOGC_VLR || logc >= _OSMO_VLR_LOGC_MAX)
+		return;
+
+	g_vlr_log_cat[logc] = logc_num;
+
+	switch (logc) {
+	case OSMO_VLR_LOGC_VLR:
+		vlr_auth_fsm_set_log_subsys(logc_num);
+		vlr_parq_fsm_set_log_subsys(logc_num);
+		vlr_lu_fsm_set_log_subsys(logc_num);
+		break;
+	case OSMO_VLR_LOGC_SGS:
+		vlr_sgs_fsm_set_log_subsys(logc_num);
+		break;
+	default:
+		break;
+	}
+}
