@@ -29,8 +29,6 @@ enum vlr_fsm_auth_event {
 	VLR_AUTH_E_MS_ID_IMSI,
 };
 
-extern struct osmo_fsm vlr_auth_fsm;
-
 struct osmo_fsm_inst *auth_fsm_start(struct vlr_subscr *vsub,
 				     struct osmo_fsm_inst *parent,
 				     uint32_t parent_event_success,
@@ -39,4 +37,6 @@ struct osmo_fsm_inst *auth_fsm_start(struct vlr_subscr *vsub,
 				     bool is_r99,
 				     bool is_utran);
 
+void vlr_auth_fsm_init(bool is_ps);
+void vlr_auth_fsm_set_log_subsys(int log_level);
 bool auth_try_reuse_tuple(struct vlr_subscr *vsub, uint8_t key_seq);
